import 'dart:ui';

import 'package:after_layout/after_layout.dart';
import 'package:animations/animations.dart';
import 'package:bolter_flutter/src/navigation/bolter_route.dart';
import 'package:bolter_flutter/src/value_stream_builder.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';

import '../bolter_provider.dart';

typedef BolterDialog = Future Function(BuildContext context);

class BolterNavigator<P extends NavigationPresenter> extends StatefulWidget {
  final Map<Type, WidgetBuilder> routes;
  final Map<Type, BolterDialog> dialogs;
  final bool isMaterial;
  final bool log;
  final bool automaticHideKeyBoardOnPush;

  const BolterNavigator({
    Key key,
    this.routes,
    this.isMaterial = false,
    this.log = false,
    this.automaticHideKeyBoardOnPush = false,
    this.dialogs,
  }) : super(key: key);

  @override
  _BolterNavigatorState createState() => _BolterNavigatorState<P>();
}

class _BolterNavigatorState<P extends NavigationPresenter>
    extends State<BolterNavigator<P>> with AfterLayoutMixin {
  final _routes = <BolterRoute<Object>>[];
  int lastKnownStackSize = 0;

  Map<Type, WidgetBuilder> get routes => widget.routes;

  Map<Type, BolterDialog> get dialogs => widget.dialogs;

  @override
  Widget build(BuildContext context) =>
      routes[context.presenter<P>().firstRoute.runtimeType](context);

  void silentPop(P presenter) {
    if (_routes.length > 1) {
      lastKnownStackSize = lastKnownStackSize - 1;
      if (_routes.length > lastKnownStackSize) {
        _routes.removeLast();
        presenter.pop();
      }
    }
  }

  PageRoute _pageRoute({WidgetBuilder widgetBuilder, RouteSettings settings}) =>
      widget.isMaterial
          ? MaterialPageRoute(builder: widgetBuilder, settings: settings)
          : CupertinoPageRoute(builder: widgetBuilder, settings: settings);

  @override
  void afterFirstLayout(BuildContext context) {
    final navigator = Navigator.of(context);
    final currentContext = navigator.overlay.context;
    final presenter = context.presenter<P>();
    final initialRoutes = presenter.currentRoutes;
    _routes.addAll(initialRoutes);
    if (initialRoutes.length > 1) {
      final others = initialRoutes.sublist(1, initialRoutes.length);
      for (final route in others) {
        final target = routes[route.runtimeType] ?? dialogs[route.runtimeType];
        if (target is! WidgetBuilder) {
          (target(currentContext) as Future).then((_) {
            route.complete();
            silentPop(presenter);
          });
        } else {
          navigator
              .push(_pageRoute(widgetBuilder: target as WidgetBuilder))
              .then((value) {
            route.complete();
            silentPop(presenter);
          });
        }
      }
    }
    lastKnownStackSize = _routes.length;
    presenter.routesStream.stream.listen((_) {
      final newRoutes = presenter.currentRoutes;
      if (widget.log) {
        // ignore: avoid_print
        print('Last routes: $_routes');
        // ignore: avoid_print
        print('New routes: $newRoutes');
      }
      final routesDifferentNumber = newRoutes.length - _routes.length;
      if (routesDifferentNumber == 0 && newRoutes.last != _routes.last) {
        final newRoute = newRoutes.last;
        final target =
            routes[newRoute.runtimeType] ?? dialogs[newRoute.runtimeType];
        if (target is! WidgetBuilder) {
          (target(currentContext) as Future).then((_) {
            newRoute.complete();
            silentPop(presenter);
          });
        } else {
          navigator
              .pushReplacement(
                  _pageRoute(widgetBuilder: target as WidgetBuilder))
              .then((value) {
            newRoute.complete();
            silentPop(presenter);
          });
        }
        lastKnownStackSize = lastKnownStackSize + 1;
      } else if (routesDifferentNumber > 0) {
        final pushRoutes = newRoutes.sublist(_routes.length, newRoutes.length);
        for (final route in pushRoutes) {
          final target =
              routes[route.runtimeType] ?? dialogs[route.runtimeType];
          if (target is! WidgetBuilder) {
            (target(currentContext) as Future).then((_) {
              route.complete();
              silentPop(presenter);
            });
          } else {
            navigator
                .push(_pageRoute(widgetBuilder: target as WidgetBuilder))
                .then((_) {
              route.complete();
              silentPop(presenter);
            });
          }
        }
        lastKnownStackSize = newRoutes.length;
      } else if (routesDifferentNumber < 0) {
        if (_routes[0] != newRoutes[0]) {
          lastKnownStackSize = newRoutes.length;
          final newRoute = newRoutes.first;
          final target =
              routes[newRoute.runtimeType] ?? dialogs[newRoute.runtimeType];
          if (target is! WidgetBuilder) throw Error();
          navigator
              .pushAndRemoveUntil(
                  _pageRoute(widgetBuilder: target as WidgetBuilder),
                  (route) => false)
              .then((value) {
            newRoute.complete();
            silentPop(presenter);
          });
        } else {
          _routes.sublist(newRoutes.length, _routes.length).forEach((route) {
            navigator.pop();
          });
        }
      }
      _routes.clear();
      _routes.addAll(newRoutes);
    });
  }
}

class BolterTabNavigator<P extends TabNavigationPresenter>
    extends StatefulWidget {
  final Map<String, WidgetBuilder> pages;
  final Map<String, Widget> tabs;
  final Map<String, Widget> selectedTabs;
  final Color tabBarColor;
  final double tabsPadding;
  final bool expandBody;
  final double blurValue;
  final bool fadeInOut;

  const BolterTabNavigator({
    Key key,
    @required this.pages,
    this.tabBarColor = Colors.transparent,
    @required this.tabs,
    @required this.selectedTabs,
    this.tabsPadding = 0.0,
    this.expandBody = false,
    this.blurValue = 0.0,
    this.fadeInOut = true,
  }) : super(key: key);

  @override
  _BolterTabNavigatorState<P> createState() => _BolterTabNavigatorState<P>();
}

class _BolterTabNavigatorState<P extends TabNavigationPresenter>
    extends State<BolterTabNavigator<P>> with AfterLayoutMixin {
  double tabBarHeight = 0.0;
  final gk = GlobalKey();

  @override
  void afterFirstLayout(BuildContext context) {
    tabBarHeight =
        (gk.currentContext.findRenderObject() as RenderBox).size.height;
  }

  @override
  Widget build(BuildContext context) {
    final presenter = context.presenter<P>();
    final mq = MediaQuery.of(context);
    final size = mq.size;
    return ValueStreamBuilder<String>(
        valueStream: presenter.currentTabStream,
        builder: (_, value) {
          final currentPage = widget.pages[value](context);
          final currentPresenter =
              currentPage is PresenterProvider<NavigationPresenter>
                  ? currentPage.presenter
                  : null;
          return OrientationBuilder(builder: (context, orientation) {
            final isPortrait = orientation == Orientation.portrait;
            final children = widget.pages.keys.map((tab) {
              return Expanded(
                key: ValueKey(tab),
                child: GestureDetector(
                  onTap: () {
                    if (value != tab) {
                      presenter.changeTab(tab);
                    } else {
                      final routes = currentPresenter?.currentRoutes ?? [];
                      if (routes.length > 1) {
                        currentPresenter.pushAndRemoveUntil(routes.first);
                      }
                    }
                  },
                  child: SafeArea(
                    top: false,
                    left: false,
                    right: false,
                    bottom: isPortrait,
                    child: Container(
                      key: tab == widget.tabs.keys.first ? gk : null,
                      color: Colors.transparent,
                      padding: isPortrait
                          ? EdgeInsets.symmetric(vertical: widget.tabsPadding)
                          : EdgeInsets.symmetric(
                              horizontal: widget.tabsPadding),
                      alignment: Alignment.center,
                      child: value == tab
                          ? widget.selectedTabs[tab]
                          : widget.tabs[tab],
                    ),
                  ),
                ),
              );
            }).toList();
            final switcher = widget.fadeInOut
                ? PageTransitionSwitcher(
                    transitionBuilder: (Widget child,
                        Animation<double> animation,
                        Animation<double> secondaryAnimation) {
                      return FadeThroughTransition(
                        animation: animation,
                        secondaryAnimation: secondaryAnimation,
                        child: child,
                      );
                    },
                    child: currentPage,
                  )
                : IndexedStack(
                    index: widget.pages.keys.toList().indexOf(value),
                    children: widget.pages.values
                        .map((value) => value(context))
                        .toList(),
                  );
            final tabBar = Container(
              decoration: BoxDecoration(
                  color: widget.tabBarColor,
                  border: const Border(
                      top: BorderSide(width: 0.5, color: Colors.black38))),
              constraints: isPortrait
                  ? BoxConstraints(maxWidth: size.width)
                  : BoxConstraints(maxHeight: size.height),
              child: orientation == Orientation.portrait
                  ? Row(
                      children: children,
                    )
                  : Column(children: children),
            );
            final portrait = widget.expandBody
                ? Stack(
                    children: [
                      switcher,
                      if (widget.blurValue != 0.0)
                        Positioned(
                          bottom: 0,
                          child: ClipRect(
                            child: BackdropFilter(
                              filter: ImageFilter.blur(sigmaX: 10, sigmaY: 10),
                              child: Container(
                                width: size.width,
                                height: tabBarHeight + mq.padding.bottom,
                                color: Colors.transparent,
                              ),
                            ),
                          ),
                        ),
                      Positioned(
                        bottom: 0,
                        child: tabBar,
                      ),
                    ],
                  )
                : Column(
                    children: [
                      Expanded(
                        child: switcher,
                      ),
                      tabBar
                    ],
                  );
            final landscape = widget.expandBody
                ? Stack(
                    children: [
                      switcher,
                      if (widget.blurValue != 0.0)
                        Positioned(
                          right: 0,
                          child: ClipRect(
                            child: BackdropFilter(
                              filter: ImageFilter.blur(sigmaX: 10, sigmaY: 10),
                              child: Container(
                                width: tabBarHeight,
                                height: size.height,
                                color: Colors.transparent,
                              ),
                            ),
                          ),
                        ),
                      Positioned(
                        right: 0,
                        child: tabBar,
                      )
                    ],
                  )
                : Column(
                    children: [
                      Expanded(
                        child: switcher,
                      ),
                      tabBar
                    ],
                  );

            return isPortrait ? portrait : landscape;
          });
        });
  }
}
